////////////////////////////////////////////////////////////////////////////////////////////////////
// Sketch for runing two stepper motors in sync. 
////////////////////////////////////////////////////////////////////////////////////////////////////

// Delete the next line to use full steps. The motor will run a bit less smooth, but faster.
#define __USE_HALF_STEPS__ 

#define AXIS_SELECT_LED_PIN 10
#define MODE_SELECT_LED_PIN 11
#define NUM_MENUS      5
#define MAX_LED_LEVEL  160
int lightLevel = 0;
int lightInc = 1;
int lightIncSign = 1;

#include <SPI.h>
#include <AccelStepper.h>

#define POS_START      0
#define POS_END        1
#define MAX_SPEED      1000
#define MAX_ACCEL      1000

#ifdef __USE_HALF_STEPS__ 
  #define STEPS_PER_ROT 4076  // For 28-BYJ48 according to most sources, half steps.
#else
  #define STEPS_PER_ROT 2038  // For 28-BYJ48 according to most sources, full steps.
#endif

AccelStepper* stepper[2];
long axisPos[2][2] = {{0, 0}, {0, 0}};
long machineSpeed = 680;
int selectedMenu = 0;
int buttonStatusOld[3] = {0, 0, 0};
int buttonStatus[3] = {0, 0, 0};;
bool machineIsMoving = false;

void executeMotion(int direction);
void moveAxisManually(int axis, long position, int direction);

////////////////////////////////////////////////////////////////////////////////////////////////////
// Initialize the conrtoller.
////////////////////////////////////////////////////////////////////////////////////////////////////
void setup(){
  int i;
  
  // Change speed by poti, not by menu; + 1 for avoiding freezes.
  machineSpeed = (long)analogRead(17) * MAX_SPEED / 1023 + 1; 

  // Initialize the internal pull-up resistors for the push buttons.
  pinMode(14, INPUT_PULLUP); // Red button. (A0 = 14 ... A2 = 16)
  pinMode(15, INPUT_PULLUP); // Yellow button.
  pinMode(16, INPUT_PULLUP); // Green button.

  // Initialize axes.
#ifdef __USE_HALF_STEPS__ 
  stepper[0] = new AccelStepper(AccelStepper::HALF4WIRE, 2, 4, 3, 5); // On ULN2003: Pin1, Pin3, Pin2, Pin4.
  stepper[1] = new AccelStepper(AccelStepper::HALF4WIRE, 6, 8, 7, 9); // On ULN2003: Pin1, Pin3, Pin2, Pin4.
#else
  stepper[0] = new AccelStepper(AccelStepper::FULL4WIRE, 2, 4, 3, 5); // On ULN2003: Pin1, Pin3, Pin2, Pin4.
  stepper[1] = new AccelStepper(AccelStepper::FULL4WIRE, 6, 8, 7, 9); // On ULN2003: Pin1, Pin3, Pin2, Pin4.
#endif
  stepper[0]->setCurrentPosition(0);
  stepper[0]->setMaxSpeed(machineSpeed); 
  stepper[0]->setAcceleration(machineSpeed);

  stepper[1]->setCurrentPosition(0);
  stepper[1]->setMaxSpeed(machineSpeed); 
  stepper[1]->setAcceleration(machineSpeed);

  // Switch off both LEDs.
  pinMode(AXIS_SELECT_LED_PIN, OUTPUT);
  pinMode(MODE_SELECT_LED_PIN, OUTPUT);
  digitalWrite(AXIS_SELECT_LED_PIN, LOW);
  digitalWrite(MODE_SELECT_LED_PIN, LOW);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Main loop.
////////////////////////////////////////////////////////////////////////////////////////////////////
void loop() {
  char textBuffer[64];
  int newStatus;
  int i;

  if(selectedMenu == 0){
    // Do "smooth blinking" if the system is in menu 0 (waiting for "execution").
    // First, adjust the blink speed according to the currently selected machine speed.
    lightInc = (analogRead(17) / 100) + 1; // Poti can go up to 1023, so this will go from 1 to 11
    analogWrite(AXIS_SELECT_LED_PIN, lightLevel);
    analogWrite(MODE_SELECT_LED_PIN, lightLevel);
    lightLevel += lightInc*lightIncSign;
    lightLevel = min(lightLevel, MAX_LED_LEVEL);
    lightLevel = max(lightLevel, 0);
    if((lightLevel == MAX_LED_LEVEL)||(lightLevel == 0)){
      lightIncSign *= -1;  
    }
  }
  // Check whether a button has been pressed.
  for(i = 0; i < 3; i++){
    buttonStatus[i] = !digitalRead(14+i); // 14 == A0; "!" because button will be "HIGH" when _not_ being pressed.
  }
  if(buttonStatus[0] && !buttonStatusOld[0] && !buttonStatus[1]&& !buttonStatus[2]){  // Menu selection button was pressed and all other buttons are not pressed.
    // Increase menu if the selection button is not blocked by an ongoing operation.
    if(!machineIsMoving){
      selectedMenu ++;
      selectedMenu %= NUM_MENUS;
      // When using LEDs, switch them on/off here.
      switch(selectedMenu){
        case 1: // Axis 1 - Start Position.
          digitalWrite(AXIS_SELECT_LED_PIN, LOW);
          digitalWrite(MODE_SELECT_LED_PIN, LOW);
          break;
        case 2: // Axis 2 - Start Position.
          digitalWrite(AXIS_SELECT_LED_PIN, HIGH);
          digitalWrite(MODE_SELECT_LED_PIN, LOW);
          break;
        case 3: // Axis 1 - End Position.
          digitalWrite(AXIS_SELECT_LED_PIN, LOW);
          digitalWrite(MODE_SELECT_LED_PIN, HIGH);
          break;
        case 4: // Axis 2 - End Position.
          digitalWrite(AXIS_SELECT_LED_PIN, HIGH);
          digitalWrite(MODE_SELECT_LED_PIN, HIGH);
          break;
      }
      goto FINALIZE_BUTTONS;
    }
  }
  // Check middle and right button.
  for(i = 1; i <= 2; i++){
    if(buttonStatus[i] != buttonStatusOld[i]){  // Go left/right or stop motion.
      if(buttonStatus[i] == 1){  // Button was pressed.
        switch(selectedMenu){
          case 0:
            executeMotion(i);
            break;
          case 1:
            // Move the selected axis to the left / right as long as the button is pressed.
            moveAxisManually(0, POS_START, i);
            break;
          case 2:
            moveAxisManually(1, POS_START, i);
            break;
          case 3:
            moveAxisManually(0, POS_END, i);
            break;
          case 4:
            moveAxisManually(1, POS_END, i);
            break;
        }
      }
    }
  }
FINALIZE_BUTTONS:    
  for(i = 0; i < 3; i++){
    buttonStatusOld[i] = buttonStatus[i];
  }
  delay(10);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Move both motors in one stroke to start (middle button) or 
// end (right button) position while monitoring the 
// left (red) button for an emergency stop.
////////////////////////////////////////////////////////////////////////////////////////////////////
void executeMotion(int direction){
  double distance[2];
  double target[2];

  target[0] = axisPos[0][direction - 1];
  target[1] = axisPos[1][direction - 1];
  // Calculate relative speed, so both drives will (approximately) stop at the same time.
  distance[0] = abs(target[0] - stepper[0]->currentPosition());
  distance[1] = abs(target[1] - stepper[1]->currentPosition());
  // The axis with the longer travel distance should use the currently defined machineSpeed.
  // The other axis must run slower. Use ceil to make sure that speed is never zero if the distance to go is very small but not zero.
  // Change speed by poti, not by menu; + 1 for avoiding freezes.
  machineSpeed = (long)analogRead(17) * MAX_SPEED / 1023 + 1; 
  if(distance[0] >= distance[1]){
    stepper[0]->setMaxSpeed(machineSpeed);    
    stepper[0]->setAcceleration(machineSpeed);
    stepper[1]->setMaxSpeed((float)(ceil(machineSpeed * distance[1]/distance[0])));    
    stepper[1]->setAcceleration((float)(ceil(machineSpeed * distance[1]/distance[0])));    
  }
  else{
    stepper[0]->setMaxSpeed((float)(ceil(machineSpeed * distance[0]/distance[1])));    
    stepper[0]->setAcceleration((float)(ceil(machineSpeed * distance[0]/distance[1])));    
    stepper[1]->setMaxSpeed(machineSpeed);    
    stepper[1]->setAcceleration(machineSpeed);    
  }
  stepper[0]->moveTo((long)target[0]);
  stepper[1]->moveTo((long)target[1]);
  machineIsMoving = true;
  digitalWrite(AXIS_SELECT_LED_PIN, HIGH);
  digitalWrite(MODE_SELECT_LED_PIN, HIGH);
  while(machineIsMoving){
    if((stepper[0]->distanceToGo() == 0)&&(stepper[1]->distanceToGo() == 0)){
      machineIsMoving = false;
    }
    stepper[0]->run();
    stepper[1]->run();
    if(!digitalRead(14)){ // Check the red button.
      machineIsMoving = false;
      stepper[0]->stop();
      stepper[1]->stop();
      stepper[0]->runToPosition(); 
      stepper[1]->runToPosition(); 
    }
  }
  digitalWrite(AXIS_SELECT_LED_PIN, LOW);
  digitalWrite(MODE_SELECT_LED_PIN, LOW);
  lightLevel = 0;
  lightInc = 1;
  waitUntilAllButtonsAreReleased();
}

////////////////////////////////////////////////////////////////////////////////
// Menu #2...4: Move the selected axis while the corresp. button is pressed.
////////////////////////////////////////////////////////////////////////////////
void moveAxisManually(int axis, long position, int direction){
  long axisLimitPosition;
  
  machineIsMoving = true;
  // Change speed by poti, not by menu; + 1 for avoiding freezes.
  machineSpeed = (long)analogRead(17) * MAX_SPEED / 1023 + 1; 
  // Travel to the start postion for the teach operation.
  // direction == 1 --> go to start, direction == 2 --> go to end.
  // The formula "(x-1)*2-1" results in -1 for x == 1 and +1 for x == 2.
  axisLimitPosition = stepper[axis]->currentPosition() + 20000 * ((direction - 1) * 2 - 1);
  // Now, go to the defined limit until the yellow/green button is released 
  /// (or the move of 20000 incfements is complete).
  stepper[axis]->moveTo(axisLimitPosition);
  stepper[axis]->setMaxSpeed(machineSpeed);    
  stepper[axis]->setAcceleration(machineSpeed);    
  while(!digitalRead(14 + direction)){ // 15 = A1 = yellow button, 16 = A2 = green button.
    // Execute the motion.
    stepper[axis]->run();
  }
  stepper[axis]->stop();
  stepper[axis]->runToPosition(); 
  axisPos[axis][position] = stepper[axis]->currentPosition();
  machineIsMoving = false;
  waitUntilAllButtonsAreReleased();
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Wait until all buttons are released.
////////////////////////////////////////////////////////////////////////////////////////////////////
void waitUntilAllButtonsAreReleased(){
  int i;

  while(!(digitalRead(14) && digitalRead(15) && digitalRead(16))){
    delay(1);  
  }
  for(i = 0; i < 3; i++){
    buttonStatusOld[i] = 0;
  }
}
